/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mouse Gestures for Mozilla.
 *
 * The Initial Developer of the Original Code is Pavol Vaskovic.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s) (alphabetical order):
 *  Andy Edmonds <aedmonds@mindspring.com>
 *  Benjamin K. Stuhl <tiriath@yahoo.com>
 *  David Illsley <illsleydc@bigfoot.com>
 *  HJ van Rantwijk <bugs4HJ@netscape.net>
 *  Jens Bannmann <jens.b@web.de>
 *  Martin.T.Kutschker <Martin.T.Kutschker@blackbox.net>
 *  Pavol Vaskovic <pali@pali.sk>
 *  Pekka Aleksi Knuutila <aleksi.knuutila@edu.lahti.fi>
 *  Scott R. Turner <srt@aero.org>
 *  Tim Williamson <chsman@hotmail.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// Global variables
// mouse gesture condition (to be evaluated on event) for more details see mozgestReadPref()
var condition = new String;
var grid; // grid size; minimal gesture has to be 'grid' pixels long
var delay; // delay before aborting gesture
var lmbGestureLimit;
var dragdropDelay;
var gestureStartTime;
var rockersEnabled;
var mgWheelRockers;
var mgWheelSensitivity;
var dTolerance; // tolerance for diagonal movement. see processCoordinates()
var gestureInProgress = false;
// old coordinates from previous gesture stroke used in processCoordinates()
var old_x;
var old_y;
var gridMoves = new Array();
var gesture = new Array();
var localizedGesture = new Array();
var mgString; // array holding localized stroke names
var lastgesture;
// lastGestureTime is time of last gesture stroke, compared with delay and current time
var lastGestureTime;
var lastMoveTime;
// timer used to abort gestures after delay, see processCoordinates()
var timer;

// trail settings
var drawTrail;
var trailLimit;
var trailWidth;
var trailInterval;
var trailColor;
var trailOpacity;

// globalOnLink is an array that holds a list of all the links traversed during gesture
// globalOnImage is a string containing an image href or false
// globalSrcEvent stores event which started the gesture the active gesture.
// The links are used in the gesture functions in gesture.js.
var globalOnLink = false;
var globalOnImage = false;
var globalSrcEvent = false;
var bundle = null; // String bundle for localized strings
var disableContextMenu = false;
var mgLastHoveredElement = null;
var mgKnownElements = null;

// used by some gesture implementations
const kWindowMediator = Components
      .classes["@mozilla.org/appshell/window-mediator;1"]
      .getService(Components.interfaces.nsIWindowMediator);

// preferences root branch, used by gesture implementations
var prefRoot = Components.classes["@mozilla.org/preferences-service;1"]
                 .getService(Components.interfaces.nsIPrefService)
                 .getBranch("");

// Preferences observer object; implements nsIObserver
function mgPrefObserver()
{
  try {
    var pbi = pref.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
    pbi.addObserver(this.domain, this, false);
  } catch(ex) {
    dump("MozGest: Failed to observe prefs: " + ex + "\n");
  }
}
mgPrefObserver.prototype = {
  domain: "mozgest.",
  observe: function(subject, topic, prefName) {
    // when MozGest pref was changed, we reinitialize
    mozgestInit();
  }
}

// This code is performed on load:
// get the preferences service
var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                            .getService(Components.interfaces.nsIPrefService);
var pref = prefService.getBranch(null); // preferences root node
var mgPref = prefService.getBranch("mozgest."); // preferences mozgest node
window.mozgestPrefObserver = new mgPrefObserver(); // and set an Pref observer
// END of on load performed code

function mozgestDisableContextMenu(aEvent) {
  if (disableContextMenu) {
    aEvent.preventDefault();
    disableContextMenu = false;
  } else {
    resetRocker();
  }
}

function mgStartup(e) {
  // we *should* remove our handler here, but that may cause a bug. 
  // see http://bugzilla.mozilla.org/show_bug.cgi?id=174320
  // instead, we stay registered and just check whether this is the correct
  // onload event.
  if (e.view != window) {
    return;
  }
  // window.removeEventListener("load", mgStartup, true);

  dump("MozGest: Startup\n");
  
  dockElement("stringbundleset", "stringbundle", "mozgestbundle",
      new Array("src"),
      new Array("chrome://mozgest/locale/mozgest.properties"));
  dockElement("keyset", "key", "mgPrefKey",
      new Array("modifiers",   "key", "command"  ),
      new Array("accel shift", "M",   "mgPrefCmd"));
  dockElement("commandset", "command", "mgPrefCmd",
      new Array("oncommand"),
      new Array("showMgPrefPanel();"));

  mozgestLoadStrings();
  determineWindowType();
  mozgestInit();
}

function mozgestInit() {
  // read prefs (or set default for each missing setting)
  mozgestReadPref();

  // turn gestures recognition on/off
  if (mgIsEnabled()) {
    addWindowWatch();
  } else {
    removeWindowWatch();
  }

  loadGestureTable();

  // context menu disabler for right mouse button
  var contentAreaContextMenu = mgGetContextMenu();
  if (contentAreaContextMenu)
    contentAreaContextMenu.addEventListener("popupshowing", mozgestDisableContextMenu, false);
}

function loadGestureTable() {
  dump("Loading gesture table\n");
  resetGesturesTable();
  mgStorage.initUserRDF();
  mgStorage.readRDFGestureTable("window");
  mgStorage.readRDFGestureTable(windowType);
}

function mozgestLoadStrings() {
  // load localized string bundle and cache often used strings
  bundle = document.getElementById("mozgestbundle");
  mgString = new Array();
  mgString["R"] = bundle.getString("abbreviation.right");
  mgString["L"] = bundle.getString("abbreviation.left");
  mgString["U"] = bundle.getString("abbreviation.up");
  mgString["D"] = bundle.getString("abbreviation.down");
  mgString["1"] = bundle.getString("abbreviation.d1");
  mgString["3"] = bundle.getString("abbreviation.d3");
  mgString["7"] = bundle.getString("abbreviation.d7");
  mgString["9"] = bundle.getString("abbreviation.d9");
  mgString["gesture"] = bundle.getString("g.gesture");
}

function processCoordinates(e){ // e is browser's event object
  var g = gesture.join("");
  if (condition == "e.button==0" && (g == "L" || g == "R")
      && (e.timeStamp - gestureStartTime) > lmbGestureLimit) {
    endGesture(null);
  }

  mozgestDrawTrail(e);

  var x_dir = e.clientX - old_x;
  var y_dir = e.clientY - old_y;
  var x = Math.abs(x_dir);
  var y = Math.abs(y_dir);

  if ( (x >= grid/2 || y >= grid/2) ) { // process each half-grid
    // diagonal movement:
    // dTolerance = 75 means that a movement is recognized as diagonal when 
    // x/y or y/x is between 0.25 and 1
    if ( dTolerance != 0 && ( (x/y >= (1-dTolerance/100) && x/y <= 1)
        || (y/x >= (1-dTolerance/100) && y/x <= 1) ) ) {

      if (x_dir < 0 && y_dir > 0) {
        gPush("1");
      } else if (x_dir > 0 && y_dir > 0) {
        gPush("3");
      } else if (x_dir < 0 && y_dir < 0) {
        gPush("7");
      } else if (x_dir > 0 && y_dir < 0) {
        gPush("9");
      }

    }
    // horizontal movement:
    else if (x > y) {
      if (x_dir > 0) {
        gPush("R");
      } else if (x_dir < 0) {
        gPush("L");
      }
    }
    // vertical movement:
    else if (x < y) {
      if (y_dir > 0){
        gPush("D");
      } else if (y_dir < 0){
        gPush("U");
      }
    }
    mgMsg(mgString["gesture"] + " " + localizedGesture);
    lastGestureTime = e.timeStamp;
    if (timer) { window.clearTimeout(timer); }
    timer = window.setTimeout("endGesture(null);", delay);

    old_x = e.clientX;
    old_y = e.clientY;
  }
  lastMoveTime = e.timeStamp;

  mgFindHoveredElements(e.originalTarget);
}

function gPush(code) {
  gridMoves.push(code);
  if (gridMoves.length == 2) {
    if (gridMoves[0] == gridMoves[1] || gesture.length == 0) {
      doPush(gridMoves.pop());
    }
    gridMoves.length = 0;
  }
}

function doPush(code) {
  if (gesture[gesture.length-1] != code) {
    gesture.push(code);
    localizedGesture.push(mgString[code]);
  }
  if (gesture.length == 1) resetRocker();
}

function mgFindHoveredElements(domPosition) {
  while (domPosition != null && domPosition != mgGetContentArea()
         && !mgKnownElements[domPosition]) {
    mgKnownElements[domPosition] = true;
    mgExamineHoveredElement(domPosition);
    domPosition = domPosition.parentNode;
  }
}

function mgExamineHoveredElement(node) {
  var tag = node.nodeName.toLowerCase();
  if (tag == "a" && node.href && !mozgestCheckLinkDups(node.href)) {
    if(!globalOnLink) globalOnLink = new Array();
    globalOnLink.push(node);
    dump("Mozgest: hovered link '" + node.href + "'\n");
  } else if (tag == "img" && globalOnImage == false) {
    globalOnImage = node;
    dump("Mozgest: found image.\n");
  }
}

function mozgestCheckLinkDups(addy) { // avoid duplicates in globalOnLink array
  var isADup = false;
  for(var zz = 0; zz < globalOnLink.length; zz++) {
    if(globalOnLink[zz].href == addy) isADup = true;
  }
  return isADup;
}

// startGesture is called from event listener in bubble phase
function startGesture(e){
  // collect ancestor elements. this is used for finding
  // initial onlink/image and inContent checks
  mgKnownElements = new Array();
  var ancestors = new Array();
  var cur = e.originalTarget;
  while (cur != null && cur != mgGetContentArea()) {
    var tag = cur.nodeName.toLowerCase();
    if (!ancestors[tag]) ancestors[tag] = cur;
    mgKnownElements[cur] = true;
    cur = cur.parentNode;
  }

  if (mozgestIsInContent(ancestors)
      && !checkForRocker(e) // note: rockers *can* start while gestureInProgress
      && !gestureInProgress
      && eval(condition)) {
    globalSrcEvent = e;
    gestureInProgress = true;
    gestureStartTime = e.timeStamp;
    lastMoveTime = 0;
    old_x = e.clientX;
    old_y = e.clientY;

    // check if we started over link or image
    if (ancestors["a"]) mgExamineHoveredElement(ancestors["a"]);
    if (ancestors["img"]) mgExamineHoveredElement(ancestors["img"]);

    gesture.length = 0; // clear gesture array
    localizedGesture.length = 0; // clear localized gesture array
    mozgestBeginTrail(e);
    window.addEventListener("mousemove", processCoordinates, false);
    lastGestureTime = e.timeStamp;
  }
}

function endGesture(e){
  releaseRocker(e);
  if (gestureInProgress) {
    window.removeEventListener("mousemove", processCoordinates, false);
    mozgestEndTrail();
    window.clearTimeout(timer);
    lastgesture = gesture.join("");

    // check if we've been called via a timeout; see processCoordinates()
    if (e == null) {
      dump("gesture timeouted\n");
      mgMsg(bundle.getString("g.aborted"));
    } else if (e == "drag") {
    } else if (lastgesture != ""){
      /* we should stop default actions here (popup menus, text selecting)
       but I think that current Mozilla implementation does those things
       internaly, so we can not stop them from DOM. */
      e.preventDefault();
      e.stopPropagation();

      // Disable context menu but only if the right mouse button
      // is configured as the trigger
      if(mgPref.getIntPref("mousebutton") == 2) disableContextMenu = true;

      fireGesture(lastgesture);
    }
    else dump("MozGest: Gesture was too small\n");

    gestureInProgress = false;
    gestureStartTime = false;
    globalOnLink = false;
    globalOnImage = false;
    globalSrcEvent = false;
  }
}

function mozgestReadPref(){ // read settings from Preferences interface
  grid = mgGetPref("grid", "Int", 15, 10, 500);
  delay = mgGetPref("delay", "Int", 500);
  lmbGestureLimit = mgGetPref("lmbGestureLimit", "Int", 500, 100, 10000);
  dragdropDelay = mgGetPref("dragdropDelay", "Int", 700, 100, 10000);
  rockersEnabled = mgGetPref("enableRockers", "Bool", true);
  mgWheelRockers = mgGetPref("wheelRockers", "Bool", false);
  mgWheelSensitivity = mgGetPref("wheelSensitivity", "Int", 1, 1, 9);
  dTolerance = mgGetPref("diagonalTolerance", "Int", 60, 0, 75);

  // condition is a boolean expression consisting of event object values
  condition = "e.button==" + mgGetPref("mousebutton", "Int", 0, 0, 2);
  if (mgGetPref("modifier.ctrl", "Bool", false))
    condition += " && e.ctrlKey";
  if (mgGetPref("modifier.alt", "Bool", false))
    condition += " && e.altKey";
  if (mgGetPref("modifier.shift", "Bool", false))
    condition += " && e.shiftKey";

  drawTrail = mgGetPref("trails.enabled", "Bool", false);
  trailWidth = mgGetPref("trails.width", "Int", 5, 1, 100);
  trailInterval = mgGetPref("trails.interval", "Int", 2, 1, 100);
  trailColor = mgGetPref("trails.color", "Char", "#FF0000");
  trailLimit = mgGetPref("trails.length-limit", "Int", 1000, 1, 10000);
  trailOpacity = mgGetPref("trails.opacity", "Int", 100, 1, 100);
}

// Gets a pref setting. If it's missing, set it to a default value.
// For Int prefs, you can optionally specify a minimum and maximum value.
function mgGetPref(name, type, defaultValue, min, max) {
  if ((type != "Bool" && type != "Char" && type != "Int") || !name) {
    return null;
  }
  var notfound = false;
  var value;
  try {
    value = eval("mgPref.get" + type + "Pref(\"" + name + "\");");
  } catch (e) {
    notfound = true;
  }
  if (notfound || (type == "Int" && (min || max) && (value > max || value < min))) {
    val = (type == "Char") ? ("\"" + defaultValue + "\"") : defaultValue;
    eval("mgPref.set" + type + "Pref(\"" + name + "\", " + val + ");");
    return defaultValue;
  } else {
    return value;
  }
}

function addWindowWatch() {
  dump("MozGest: ---- Adding Listeners ---\n");
  var content = mgGetContentArea();
  content.addEventListener("mousedown", startGesture, false);
  content.addEventListener("draggesture", onDragGestureEvent, true);
  if (mgWheelRockers) {
    content.addEventListener("DOMMouseScroll", mgMousewheelHandler, false);
  }
  window.addEventListener("mouseup", endGesture, false);
}

function removeWindowWatch() {
  dump("MozGest: ---- Removing Listeners ---\n");
  var content = mgGetContentArea();
  content.removeEventListener("mousedown", startGesture, false);
  content.removeEventListener("draggesture", onDragGestureEvent, true);
  content.removeEventListener("DOMMouseScroll", mgMousewheelHandler, false)
  window.removeEventListener("mouseup", endGesture, false);
}

function onDragGestureEvent(e) {
  var draggedElement = e.originalTarget.nodeName.toUpperCase();
  if (gestureInProgress && draggedElement != "HTML") {
    if (lastMoveTime-gestureStartTime > dragdropDelay) {
      dump("MozGest: '"+draggedElement+"' drag encountered, canceling gesture\n");
      endGesture("drag");
    } else {
      dump("MozGest: canceling '"+draggedElement+"' drag event, continuing gesture\n");
      e.stopPropagation();
    }
  }
}

function showMgPrefPanel() {
  window.openDialog("chrome://mozgest/content/pref/mozgestPrefWindow.xul",
                    "mozgestprefs", "chrome,centerscreen,modal");
}

function dockElement(setType, type, id, aNames, aValues) {
  dockElementAt(setType, type, id, aNames, aValues, document.documentElement);
}

function dockElementAt(setType, type, id, aNames, aValues, container) {
  var sets = document.getElementsByTagName(setType);
  var set;
  if (sets && sets.length > 0) {
    set = sets[0];
  } else {
    set = document.createElement(setType);
    container.appendChild(set);
  }
  var element = document.createElement(type);
  element.setAttribute("id", id);
  for (i=0; i < aNames.length; i++) {
    element.setAttribute(aNames[i], aValues[i]);
  }
  set.appendChild(element);
}
