/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mouse Gestures for Mozilla.
 *
 * The Initial Developer of the Original Code is Jens Bannmann.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s) (alphabetical order):
 *  Jens Bannmann <jens.b@web.de>
 *  Neil <neil@parkwaycc.co.uk>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var gestureType;
var gestureInput;
var codeBox;
var appBox;
var wheelBtn;
var wheelDir;
var functionType;
var bundle;

var resourceURI;
var code;
var fPrefix;
var wType;

function initEditor() {
  gestureType = document.getElementById("gestureType");
  gestureInput = document.getElementById("gestureInput");
  codeBox = document.getElementById("code");
  appBox = document.getElementById("appearance");
  wheelBtn = document.getElementById("wheelRockerButton");
  wheelDir = document.getElementById("wheelDirection");
  functionType = document.getElementById("functionType");
  functionsTree = document.getElementById("functionsTree");
  bundle = document.getElementById("mozgestbundle");

  // get passed arguments
  resourceURI = window.arguments[0];
  code = window.arguments[1];
  var appearance = window.arguments[2];
  var selectedFunc = window.arguments[3];
  var name = window.arguments[4];
  var customCode = window.arguments[5];
  fPrefix = window.arguments[6];
  wType = window.arguments[7];

  // determine gesture type
  if (code.substr(0, 1) == ":") {
    var a = code.substr(1, 1);
    var b = code.substr(2, 1);
    if (!isNaN(b)) {
      gestureType.selectedIndex = 1;
      document.getElementById("rocker1").value = a;
      document.getElementById("rocker2").value = b;
    } else {
      gestureType.selectedIndex = 2;
      wheelBtn.value = a;
      wheelDir.selectedIndex = (b == "+") ? 1 : 0;
    }
  } else {
    codeBox.value = mgMappingLocalizer.localize(code);
    appBox.value = mgMappingLocalizer.localize(appearance);
  }
  updateTypeDeck();

  // fill function list
  var ftChildren = document.getElementById("ftChildren");
  var dsFunc = mgStorage.RDF.GetDataSourceBlocking("chrome://mozgest/content/functions.rdf");
  functionList = Components.classes['@mozilla.org/rdf/container;1']
                           .createInstance()
                           .QueryInterface(Components.interfaces.nsIRDFContainer);
  try {
    functionList.Init(dsFunc, mgStorage.RDF.GetResource("urn:mozgest:functions"));
  } catch (e) {
    dump("MozGest: Could not init functionList. " + e + "\n");
  }
  var ftRows = 0;
  var selFuncRow = -1;
  var funcEnum = functionList.GetElements();
  while (funcEnum.hasMoreElements()) {
    var func = funcEnum.getNext()
               .QueryInterface(Components.interfaces.nsIRDFResource);
    var fName = mgStorage.getProperty(dsFunc, func, "func")
    var fSupports = ","+mgStorage.getProperty(dsFunc, func, "supports")+",";
    if (fName.indexOf("mg" + fPrefix + "_") == 0
        || fName.indexOf("mgW_") == 0
        || fSupports.indexOf(","+wType+",") != -1) {
      addTreeRow(ftChildren, new Array(fName, bundle.getString(fName)));
      if (fName == selectedFunc) {
        selFuncRow = ftRows;
      }
      ftRows++;
    }
  }

  // switch to custom mode or show selected function
  if (selFuncRow >= 0) {
    functionsTree.contentView.selection.select(selFuncRow);
  } else if (customCode) { // must be non-null, non-empty string
    document.getElementById("custom").value = customCode;
    document.getElementById("mappingName").value = name;
    functionType.selectedIndex = 1;
  }
  updateFunctionDeck();
}

function updateTypeDeck() {
  document.getElementById("gestureTypeDeck").selectedIndex = gestureType
                                                             .selectedIndex;
}

function updateFunctionDeck() {
  document.getElementById("functionDeck").selectedIndex = functionType
                                                          .selectedIndex;
}

function updateSecondRocker() {
  var first = document.getElementById("rocker1").value;
  for (i=0; i < 3; i++) {
    var si = document.getElementById("second"+i);
    if (i==first) {
      si.setAttribute("disabled", true);
    } else {
      si.removeAttribute("disabled");
    }
  }
  var r2 = document.getElementById("rocker2");
  if (r2.value == first && first != -1) {
    r2.value = -1;
  }
}

function saveMapping() {
  // ignore enter-keypresses when in textarea (= multiline textbox)
  // so user can enter newlines.
  // needed for seamonkey - obsolete for fb0.6, but doesn't hurt
  if (document.commandDispatcher.focusedElement instanceof
      Components.interfaces.nsIDOMHTMLTextAreaElement) {
    return false;
  }

  var code;
  var appearance = "";
  if (gestureType.selectedIndex == 0) {
    code = mgMappingLocalizer.deLocalize(codeBox.value);
    appearance = mgMappingLocalizer.deLocalize(appBox.value);
    if (!code.match(/^\*{0,1}[1379DLUR]+$/)) {
      alert(bundle.getString("enterGesture"));
      return false;
    }
    if (appearance != "") {
      var compacted = appearance.substr(0, 1);
      for(i = 1; i < appearance.length; i++) {
        var c = appearance.substr(i, 1);
        if (c != compacted.substr(compacted.length-1, 1)) {
          compacted += c;
        }
      }
      if (compacted != code) {
        alert(bundle.getString("invalidAppearance"));
        return false;
      }
    }

  } else if (gestureType.selectedIndex == 1) {
    var a = document.getElementById("rocker1").value;
    var b = document.getElementById("rocker2").value;
    if (a == -1 || b == -1) {
      alert(bundle.getString("chooseRockerButtons"));
      return false;
    }
    code = ":" + a + b;
  } else if (gestureType.selectedIndex == 2) {
    var wb = wheelBtn.value;
    var wd = wheelDir.selectedIndex;
    if (wd == -1 || wb == -1) {
      alert(bundle.getString("chooseWheelDirBtn"));
      return false;
    }
    code = ":" +  wb + ((wd == 1) ? "+" : "-");
  }

  var name;
  var func = "";
  var custom = "";
  if (functionType.selectedIndex == 1) {
    name = document.getElementById("mappingName").value;
    if (name == "") {
      alert(bundle.getString("enterMappingName"));
      return false;
    }
    custom = document.getElementById("custom").value;
  } else {
    var row = functionsTree.contentView.selection.currentIndex;
    if (row == -1) {
      alert(bundle.getString("chooseFunction"));
      return false;
    }
    func = functionsTree.treeBoxObject.view.getCellText(row, "ftFuncCol");
    name = bundle.getString(func);
  }
  return window.opener.receiveEditedMapping(resourceURI, code, appearance,
                                            func, name, custom);
}

function recognizeGesture() {
  window.openDialog("chrome://mozgest/content/pref/recognize.xul",
                    "mozgest:recognize", "chrome,centerscreen,modal=yes");
}

function receiveGesture(code) {
  codeBox.value = mgMappingLocalizer.localize(code);
  appBox.value = "";
}
